import sys
import os.path
libdir = os.path.dirname(__file__)
sys.path.append(os.path.split(libdir)[0])
from Map import MapGenerator as mg

import heapq
import numpy as np 
from random import randint
import Visualize.Visualize as vis


class PathGenerator:
    """
    This generator calculates a path from a start coordiante to an end coordiante
    while also navigating around obsicales using an A* algoritm. This algoritm has some
    performance issues when map is bigger than 10x 10y 10z.

    This class is used to simulate a path for training the ML model. 
    
    Note: This algorithm introduces no randomization, equal maps will produce equal path.

    Example usage:

    pg = PathGenerator()

    map_generator = mg.MapGenerator(10, 10, 3, 1)

    empty_map = map_generator.create_empty_map_3d(0)

    map = pg.add_obstacles(empty_map, [[[2, 3, 0], [2, 8, 0]], [[8, 0, 0], [8, 6, 0]]])

    path = pg.find_path(map, (0, 5, 0), (9, 5, 2))
    """


    class Node():
        """ The node class"""

        def __init__(self, parent=None, position=None):
            self.parent = parent
            self.position = position

            self.g = 0
            self.h = 0
            self.f = 0

        def __eq__(self, other):
            return self.position == other.position


    def add_obstacles(self, map, obstacles):
        """
        Adds obstacles to map

        :param map: The empty map.

        :param obstacles: Array with start and stop coordinates for obstacles.

        :return: Map with obstacles
        """
        for o in range(len(obstacles)):
            for x in range(map.shape[0]):
                for y in range(map.shape[1]):
                    for z in range(map.shape[2]):
                        if [x, y, 0] >= obstacles[o][0] and [x, y, 0] <= obstacles[o][1]:
                            map[x, y, z] = 1

        return map


    def astar(self, map, start, goal):
        """
        This function runs the A* algoritm.

        :param map: 3D list that represents the map. If the map does not contain obsicales then it is filled with only 0's. 1's represent an obsticale.

        :param start: [x, y, z] representing the start position.

        :param end: [x, y, z] representing the end position.

        :returns: Optimal path from start to end.
        """
        def heuristic(a, b):
            return np.sqrt((b[0] - a[0]) ** 2 + (b[1] - a[1]) ** 2 + (b[2] - a[2]) ** 2)

        neighbours = [
                    (-1, 1, 1), (0, 1, 1), (1, 1, 1), (-1, 0, 1), (0, 0, 1), (1, 0, 1), (-1, -1, 1), (0, -1, 1), (1, -1, 1),
                    (-1, 1, 0), (0, 1, 0), (1, 1, 0), (-1, 0, 0), (1, 0, 0), (-1, -1, 0), (0, -1, 0), (1, -1, 0),
                    (-1, 1, -1), (0, 1, -1), (1, 1, -1), (-1, 0, -1), (0, 0, -1), (1, 0, -1), (-1, -1, -1), (0, -1, -1), (1, -1, -1)
                ]
        close_set = set()
        came_from = {}
        gscore = {start:0}
        fscore = {start:heuristic(start, goal)}
        oheap = []
        heapq.heappush(oheap, (fscore[start], start))

        while oheap:
            current = heapq.heappop(oheap)[1]
            if current == goal:
                data = []
                while current in came_from:
                    data.append(current)
                    current = came_from[current]
                return data
            close_set.add(current)
            print(current)

            print(len(oheap))
            for x, y, z in neighbours:
                neighbor = current[0] + x, current[1] + y, current[2] + z            
                tentative_g_score = gscore[current] + heuristic(current, neighbor)
                if 0 <= neighbor[0] < map.shape[0]:
                    if 0 <= neighbor[1] < map.shape[1]:
                        if 0 <= neighbor[2] < map.shape[2]:                 
                            if map[neighbor[0]][neighbor[1]][neighbor[2]] == 1:
                                continue
                        else:
                            # array bound z walls
                            continue
                    else:
                        # array bound y walls
                        continue
                else:
                    # array bound x walls
                    continue
                    
                if neighbor in close_set and tentative_g_score >= gscore.get(neighbor, 0):
                    continue
                    
                if  tentative_g_score < gscore.get(neighbor, 0) or neighbor not in [i[1]for i in oheap]:
                    came_from[neighbor] = current
                    gscore[neighbor] = tentative_g_score
                    fscore[neighbor] = tentative_g_score + heuristic(neighbor, goal)
                    heapq.heappush(oheap, (fscore[neighbor], neighbor))

    def player_paths(self, path, n_players):
        """
        This method takes the path and number of players and merges players to the path. 
        It also contains a variable that simulates a stop after a certain amount of timesteps.
        This will add some randomization and simulates a variance in the player progression.

        :param path: The path given by A*.

        :param n_players: Number of players to be simulated along the given path.

        :returns: Each players path on a given timestep.
        """
        
        indexes = np.zeros([n_players], dtype=int)
        stop_counters = []
        for i in range(n_players):
            stop_counters.append(int(10 * (i+1)))
        wait_time = np.zeros([n_players])
        paths = []
        not_done = True

        counter = 0
        while not_done:
            positions = []
            for i in range(n_players):
                if len(paths) > 0:
                    if path[-1] == paths[-1][i]:
                        not_done = False
                        break
                if stop_counters[i] == 0:
                    if wait_time[i] == 5:
                        stop_counters[i] = int(10 * (i+1))
                        wait_time[i] = 0
                    else:
                        wait_time[i] += 1
                else:
                    stop_counters[i] -= 1
                    indexes[i] += 1
                positions.append(path[indexes[i]])
            paths.append(positions)
            counter += 1

        return paths
        
if __name__ == "__main__":
    print("Should be called in a class and not be run in standalone.")
 
