import sys
import os.path
libdir = os.path.dirname(__file__)
sys.path.append(os.path.split(libdir)[0])

import numpy as np

class InterestPoint:
    """
    This algorithm calculates which points are of intrest given a match.
    
    It returns a path that consist of most intresting point in each given timestep. This
    path will jump around and represent which points the camera should focus on for a given timestep.
    
    Example usage:

    int_point = IP.InterestPoint([0.98, 0.50, 0.82])

    x_train = int_point.get_most_interesting_positions(matches)
    """

    def __init__(self, finish):
        self.finish = finish
        self.counter = 0

    def get_most_interesting_positions(self, matches):
        """ 
        Find most intresting points for one or multiple matches.

        :param matchs: a set of matches

        :return: Most intresting path for the given matches.
        """

        paths = []
        for match in matches:
            path = np.empty([match.shape[0], 3])

            for i in range(len(match)):
                highest_scoring_player_index = 0
                highest_score = self.__calculateScore(match[i][0])
                for j in range(len(match[i])):
                    current_score = self.__calculateScore(match[i][j])
                    if current_score > highest_score:
                        highest_score = current_score
                        highest_scoring_player_index = j
                path[i] = match[i][highest_scoring_player_index][:3]
            paths.append(path)
        return paths

    def get_most_interesting_positions_old(self, matches):
        """ 
        Find most intresting points for one or multiple matches

        :param matches: a set of matches

        :return: Most intresting path for the given matches.

        Note: Old version, not in use
        """

        paths = []
        for match in matches:
            path = np.empty([match.shape[0], match.shape[1], match.shape[2]])
            for i in range(len(match)):
                highest_scoring_player_index = 0
                highest_score = self.__calculateScore(match[i][0])
                for j in range(len(match[i])):
                    current_score = self.__calculateScore(match[i][j])
                    if current_score > highest_score:
                        highest_score = current_score
                        highest_scoring_player_index = j
                path[i] = match[i][highest_scoring_player_index]
            paths.append(path)
        return paths


#gets the most interesting positions in matches
#when there is a setback it will put it as an interisting point $steps timesteps before it happens
#currently does not handle overlapping setbacks
    def get_most_interesting_positions_alt(self, matches, steps):
        """ 
        Find most intresting points for one or multiple matches, when there is a setback then it will 
        put it as an intrest points :steps: timesteps before it happens.

        :param matchs: a set of matches
        :param steps: n-steps before

        :return: Most intresting path for the given matches.

        Note: Does not handle overlapping setbacks
        """

        target = []
        for match in matches:
            setbacks = self.__find_setbacks(match)
            positions = np.zeros([match.shape[0], 4])

            #puts the position of the player closest to finish as the POI for each timestep
            for i in range(len(positions)):
                closest_distance = 1 - self.__euclidean_distance(match[i][0][:3], self.finish)
                index = 0
                for j in range(len(match[i])):
                    distance = 1 - self.__euclidean_distance(match[i][j][:3], self.finish)
                    if distance < closest_distance:
                        closest_distance = distance
                        index = j
                positions[i] = match[i][index]

            for pos in setbacks:
                start = int(pos[0]) - steps
                local_step = steps

                #if start is less than 0 it needs to be set to 0 - and local_step would have to change accordingly
                if start < 0:
                    local_step += start
                    start = 0

                for i in range(start, start + local_step + 1):
                    positions[i] = pos[1]

            #positions = np.reshape(positions, [positions.shape[0], 1, positions.shape[1]])
            target.append(positions)
        return target




        #for match in matches
        #loop to find setbacks
        #loop trough match
        #put setbacks as interestpoint for $steps timesteps before it happens - maybe $steps after too?
        #else put the player closest to finish

    def __find_setbacks(self, match):
        """ 
        Puts setbacks as intrespoints :steps: timesteps before it happens, if no player is in setback
        then player closest to goal is selected.

        :param state: a given state

        :return: List of positions with setback
        """

        setbacks = []
        setback_last_step = False
        for i in range(len(match)):
            player_set_back = False
            for player in match[i]:
                if player[3] == 1.0:
                    if not setback_last_step and not player_set_back:
                        pos = [i, np.array([player[0], player[1], player[2], player[3]])]
                        setbacks.append(pos)
                        setback_last_step = True
                    player_set_back = True
            if not player_set_back:
                setback_last_step = False

        return setbacks

                    

    def __calculateScore(self, state):
        """ 
        Uses eucledian distance to calculate a score, the highest score will be set as the intrest point
        for this state. Setbacks is weighted very high. If two players are in setback simultaneously then the player
        that is closest to goal is deemed more intresting.

        :param state: a given player state

        :return: score for a given player state
        """
        
        not_setback = 0.0
        in_setback = False
        if state[3] != not_setback:
            in_setback = True
            self.counter += 1

        distance = 1 - self.__euclidean_distance(state[:3], self.finish) #[0.99943105, 0.34999687, 0.45879413]
        if in_setback:
            distance += 1.0

        return distance


    def __euclidean_distance(self, from_coordinate, to_coordinate):
        return np.sqrt(
            ((from_coordinate[0] - to_coordinate[0])**2) + 
            ((from_coordinate[1] - to_coordinate[1])**2) + 
            ((from_coordinate[2] - to_coordinate[2])**2)
            )


if __name__ == "__main__":
    print("Should be called in a class and not be run in standalone.")
