import sys
import os.path
libdir = os.path.dirname(__file__)
sys.path.append(os.path.split(libdir)[0])
from Map import MapGenerator as mg

import numpy as np 
from random import randint
import Visualize.Visualize as vis

class PathGenerator:
    """
    Example Usage:

    pg = PathGenerator(10, 10, 10, 1)

    map = pg.create_heatmap([9, 5, 5])

    path = pg.generate_path(map, 3, [1, 5, 0], [9, 5, 5])

    """

    def __init__(self, x, y, z, density):
        self.density = density
        map_generator = mg.MapGenerator(x, y, z, density)
        self.map = map_generator.create_empty_map_3d(x)


    def create_heatmap(self, end_coordinate):
        """ Creates a 3D heatmap that is needed to calculate the path of the players.

        Dimentions of the map is specified when creating the PathGenerator object.

        The values of each tile is calculated using Euclidean distance.

        :param end_coordinate: is the position for where players want to move.

        :return: Heatmap.
        """

        map = self.map
        for i in range(map.shape[0]):
            for j in range(map.shape[1]):
                for k in range(map.shape[2]):
                    map[i, j, k] = round(self.euclidean_distance([i, j, k], end_coordinate), 1)
                        
        return map




    def euclidean_distance(self, from_coordinate, to_coordinate):
        return np.sqrt(
            ((from_coordinate[0] - to_coordinate[0])**2) + 
            ((from_coordinate[1] - to_coordinate[1])**2) + 
            ((from_coordinate[2] - to_coordinate[2])**2)
            )

    def generate_path(self, map, player_amount, start_coordinates, end_coordinates):
        """ Generates the path for n players.

        :param player_amount: specifies the amount of players for which a path is generated.

        :param start_coordinate: is the initial position of all players. They all start from this position.

        :param end_coordinate: is the end. When all players reaches this position the calculation is complete.

        :return: All the states the players use to move from initial position to the end.
        """

        states = []
        running = True
        while(running):
            state = []
            for i in range(player_amount):
                if len(states) == 0 and len(state) < player_amount:
                    state.append(self.next_position(map, start_coordinates))
                else:
                    if states[len(states) - 1][i] != end_coordinates:
                        state.append(self.next_position(map, states[len(states) - 1][i]))
                    else:
                        state.append(states[len(states) - 1][i])
            states.append(state)
            counter = 0
            for state in states[len(states) - 1]:
                if state == end_coordinates:
                    counter += 1
                    running = False
            if counter == player_amount:
                running = False

        return states
    
    
    def next_position(self, map, player_location):
        """ Calculates next position for the player.

        :param map: is the heatmap so that it can find appropriate tiles next to the player.

        :param player_location: is where the player currently is positioned.

        :return: Next position but is polluted with noise to generate a more diverse pattern.
        """
        
        def myFunc(e):
            return e['value']
        tiles = []
        unavailable_tile = map.shape[0] * map.shape[1] * map.shape[2]

        if player_location[0] + 1 < map.shape[0]:
            forward = map[player_location[0] + 1, player_location[1], player_location[2]]
        else:
            forward = unavailable_tile
        tiles.append({'dir': 'forward', 'value': forward})

        if player_location[0] - 1 >= 0:
            backward = map[player_location[0] - 1, player_location[1], player_location[2]]
        else:
            backward = unavailable_tile
        tiles.append({'dir': 'backward', 'value': backward})

        if player_location[1] + 1 < map.shape[1]:
            left = map[player_location[0], player_location[1] + 1, player_location[2]]
        else:
            left = unavailable_tile
        tiles.append({'dir': 'left', 'value': left})
        
        if player_location[1] - 1 >= 0:
            right = map[player_location[0], player_location[1] - 1, player_location[2]]
        else:
            right = unavailable_tile
        tiles.append({'dir': 'right', 'value': right})

        if player_location[2] + 1 < map.shape[2]:
            up = map[player_location[0], player_location[1], player_location[2] + 1]
        else:
            up = unavailable_tile
        tiles.append({'dir': 'up', 'value': up})

        if player_location[2] - 1 >= 0:
            down = map[player_location[0], player_location[1], player_location[2] - 1]
        else:
            down = unavailable_tile
        tiles.append({'dir': 'down', 'value': down})

        tiles.sort(key=myFunc)
        noise = randint(0, 9)

        if noise < 3:
            if tiles[0]['dir'] == 'forward':
                return [player_location[0] + 1, player_location[1], player_location[2]]
            elif tiles[0]['dir'] == 'backward':
                return [player_location[0] - 1, player_location[1], player_location[2]]
            elif tiles[0]['dir'] == 'left':
                return [player_location[0], player_location[1] + 1, player_location[2]]
            elif tiles[0]['dir'] == 'right':
                return [player_location[0], player_location[1] - 1, player_location[2]]
            elif tiles[0]['dir'] == 'up':
                return [player_location[0], player_location[1], player_location[2] + 1]
            elif tiles[0]['dir'] == 'down':
                return [player_location[0], player_location[1], player_location[2] - 1]
        elif noise < 5:
            if tiles[1]['dir'] == 'forward':
                return [player_location[0] + 1, player_location[1], player_location[2]]
            elif tiles[1]['dir'] == 'backward':
                return [player_location[0] - 1, player_location[1], player_location[2]]
            elif tiles[1]['dir'] == 'left':
                return [player_location[0], player_location[1] + 1, player_location[2]]
            elif tiles[1]['dir'] == 'right':
                return [player_location[0], player_location[1] - 1, player_location[2]]
            elif tiles[1]['dir'] == 'up':
                return [player_location[0], player_location[1], player_location[2] + 1]
            elif tiles[1]['dir'] == 'down':
                return [player_location[0], player_location[1], player_location[2] - 1]
        else:
            if tiles[2]['dir'] == 'forward':
                return [player_location[0] + 1, player_location[1], player_location[2]]
            elif tiles[2]['dir'] == 'backward':
                return [player_location[0] - 1, player_location[1], player_location[2]]
            elif tiles[2]['dir'] == 'left':
                return [player_location[0], player_location[1] + 1, player_location[2]]
            elif tiles[2]['dir'] == 'right':
                return [player_location[0], player_location[1] - 1, player_location[2]]
            elif tiles[2]['dir'] == 'up':
                return [player_location[0], player_location[1], player_location[2] + 1]
            elif tiles[2]['dir'] == 'down':
                return [player_location[0], player_location[1], player_location[2] - 1]


if __name__ == "__main__":
    print("Should be called in a class and not be run in standalone.")
    
    
    


                    

