import os
os.environ['CUDA_LAUNCH_BLOCKING'] = "1"

import sys
import os.path
libdir = os.path.dirname(__file__)
sys.path.append(os.path.split(libdir)[0])

import Data.DataFormatter as df

import torch
import torch.nn.functional as F
import torch.nn as nn
import torch.optim as optim
import numpy as np
import Data.InterestPoint as ip
from datetime import datetime
import random

"""
A Model that tries to identify points of interest given the positions of players at the same timestep
Uses convolutions and max-pooling to decrease the resolution of the map.
Using CUDA will make it significantly faster
"""


class ConvModel(nn.Module):
    def __init__(self, factor, shape=(8070, 3780)):
        super(ConvModel, self).__init__()

        #scales the shape by the given factor
        shape = (shape[0]/factor, shape[1]/factor)

        #sets the seed for random initialization
        seed = 42
        np.random.seed(seed)
        torch.manual_seed(seed)

        self.conv1 = nn.Conv2d(1, 6, kernel_size=10, stride=5, padding=0)
        self.conv2 = nn.Conv2d(6, 12, kernel_size=5, stride=3, padding=0)
        self.pool = nn.MaxPool2d(kernel_size=5, stride=3, padding=0)

        self.output_size1 = output_size(output_size(output_size(output_size(shape[0], 10, 5, 0), 5, 3, 0), 5, 3, 0), 5, 3, 0)
        self.output_size2 = output_size(output_size(output_size(output_size(shape[1], 10, 5, 0), 5, 3, 0), 5, 3, 0), 5, 3, 0)
        self.output_size = self.output_size1*self.output_size2*12

        self.fc1 = nn.Linear(self.output_size, 468)
        self.fc2 = nn.Linear(468, 200)
        self.softmax = nn.Softmax()




    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = self.pool(x)
        x = F.relu(self.conv2(x))
        x = self.pool(x)

        x = x.view(-1, x.shape[1] * x.shape[2] * x.shape[3])
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        return x

#The main training loop
def run(factor):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    x_train, x_test, y_train, y_test = get_data(factor)
    model = ConvModel(factor)
    model.to(device)
    loss_func = torch.nn.MSELoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001)
    date = datetime.now()

    lowest_loss = 10000000.0

    for epoch in range(15):
        print("epoch ", epoch)
        print(datetime.now().time())
        running_loss = 0.0

        for i in range(len(x_train)):
            data = x_train[i]
            targets = y_train[i]
            print("map ", i, " of ", len(x_train))

            for j in range(len(data)):
                if (j%1000) == 0:
                    print("timestep ", j, " of ", len(data))
                optimizer.zero_grad()

                map_in = to_grid_2d(data[j], factor).to(device)
                target = to_grid_target(targets[j], factor).to(device)

                output = model(map_in)

                loss = loss_func(output[0], target[0])
                running_loss += loss.item()

                loss.backward()
                optimizer.step()

                del map_in, target, output, loss
            del data, targets

        print("loss: ", running_loss)

        if running_loss < lowest_loss:
            lowest_loss = running_loss
            torch.save(model.state_dict(), "Data/Models/conv_linear_" + str(date) + ".pt")
            print("saved")



#Gets the data specified by dir, scaled by $factor
#also gets the most interesting positions for both train and test
def get_data(factor):
    x_train, x_test = df.get_data_from_dir("Data/Datasets/set1", 2, factor=factor)
    int_point = ip.InterestPoint([1980, 470, 210])
    y_train = int_point.get_most_interesting_positions(x_train)
    y_test = int_point.get_most_interesting_positions(x_test)
    return x_train, x_test, y_train, y_test


#gets the output_size of either a convolution or a maxpool given the parameters
def output_size(in_size, kernel_size, stride, padding):
    return int((in_size - kernel_size + 2*padding)/stride) + 1


#creates a grid with the size of shape scaled by factor, and initializes all elements to 0
#sets the elements that corresponds to the x and y values in data to 1
def to_grid_2d(data, factor, shape=(8070, 3780)):
    shape = (int(shape[0]/factor), int(shape[1]/factor))
    grid = torch.zeros(shape)
    for x in data:
        grid[int(x[0])][int(x[1])] = 1.0
    grid = grid.view(1, 1, shape[0], shape[1])
    return grid


#Creates a grid with the size of target_shape + 1, which should correspond to the output of the network, and initializes all elements to 0
#puts the x and y positions from data, scaled from (shape/factor) to target_shape, as 1 in the grid
def to_grid_target(data, factor, shape=(8070, 3780), target_shape=(19, 9)):
    shape = (shape[0]/factor, shape[1]/factor)
    x = int((data[0][0] / shape[0]) * target_shape[0])
    y = int((data[0][1] / shape[1]) * target_shape[1])
    grid = torch.zeros(target_shape[0]+1, target_shape[1]+1)
    grid[x][y] = 1.0
    grid = grid.view(1, -1)

    for i in range(list(grid[0].shape)[0]):
        if grid[0][i] != 1.0:
            grid[0][i] = random.uniform(0.0, 0.2)
    return grid


if __name__ == "__main__":
    run(4)

