import sys
import os.path
libdir = os.path.dirname(__file__)
sys.path.append(os.path.split(libdir)[0])


import torch
import torch.nn as nn
import torch.optim as optim
import Data.DataFormatter as df
import numpy as np

import Visualize.Visualize as vis
from datetime import datetime
import Data.InterestPoint as ip

"""
Uses GRU to predict the coordinates of interesting positions directly
"""

class GRUmodel(nn.Module):
    def __init__(self, input_size, hidden_size, device):
        super(GRUmodel, self).__init__()
        seed = 17
        np.random.seed(seed)
        torch.manual_seed(seed)

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.device = device

        self.gru = nn.GRU(input_size, hidden_size)
        self.linear = nn.Linear(hidden_size, input_size)
        self.leakyRelu = torch.nn.LeakyReLU()


        self.hidden = self.init_hidden()

    def init_hidden(self):
        return torch.randn(1, 1, self.hidden_size, device=self.device)#, torch.randn(1, 1, self.hidden_size, device=self.device)


    def forward(self, input):
        out, hidden = self.gru(input.view(1, 1, -1), self.hidden)
        out = self.leakyRelu(out)
        out = self.linear(out)
        return out


#Gets the data from directory scaled to [0, 1], and gets the corresponding most interesting positions
#reshapes the numpy arrays for each match to [timesteps, features] and converts to tensors
def get_data(n_players):
    x_train, x_test = df.get_data_from_dir("Data/Datasets/larger_set", n_players, scale=True)
    int_point = ip.InterestPoint([0.98, 0.50, 0.82])
    y_train = int_point.get_most_interesting_positions(x_train)
    y_test = int_point.get_most_interesting_positions(x_test)

    for i in range(len(x_train)):
        x_train[i] = np.reshape(x_train[i], [x_train[i].shape[0], -1])
        x_train[i] = torch.from_numpy(x_train[i]).float()
        y_train[i] = np.reshape(y_train[i], [y_train[i].shape[0], -1])
        y_train[i] = torch.from_numpy(y_train[i]).float()


    for i in range(len(x_test)):
        x_test[i] = np.reshape(x_test[i], [x_test[i].shape[0], -1])
        x_test[i] = torch.from_numpy(x_test[i]).float()
        y_test[i] = np.reshape(y_test[i], [y_test[i].shape[0], -1])
        y_test[i] = torch.from_numpy(y_test[i]).float()

    return x_train, y_train, x_test, y_test


#the main training loop
def run(step_size, hidden_size, n_players):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    x_train, y_train, x_test, y_test = get_data(n_players)
    model = GRUmodel(x_train[0].shape[1], hidden_size, device)
    model.to(device)

    date = datetime.now()

    loss_func = nn.MSELoss()
    optimizer = optim.Adam(model.parameters(), lr=0.0001)
    lowest_loss = 10000000.0
    scheduler = optim.lr_scheduler.MultiStepLR(optimizer, [100, 1000, 2000], gamma=0.1)


    for epoch in range(100):
        print("epoch %i" % epoch)
        print(datetime.now().time())
        running_loss = 0.0
        for i in range(len(x_train)):
            data_in = x_train[i].to(device)
            data_out = y_train[i].to(device)

            for j in range(len(data_in)-step_size):
                optimizer.zero_grad()
                model.hidden = model.init_hidden()

                map_in = data_in[j]
                target = data_out[j+step_size]

                output = model(map_in)
                loss = loss_func(output, target)
                running_loss += loss

                loss.backward()
                optimizer.step()
        scheduler.step()

        print("loss: %f" % running_loss)
        if running_loss < lowest_loss:
            lowest_loss = running_loss
            torch.save(model.state_dict(), "Data/Models/GRU/testing" + str(date) + ".pt")
            print("saved")

    return test(x_test, y_test, model, 0, step_size, device)

#Evaluates the model on the specified map in the testset with the given step_size
def test(x_test, y_test, model, map_nr, step_size, device):
    with torch.no_grad():
        predictions = np.empty([x_test[map_nr].shape[0]-step_size, x_test[map_nr].shape[1]])
        targets = np.empty([x_test[map_nr].shape[0]-step_size, x_test[map_nr].shape[1]])
        test = x_test[map_nr].to(device)

        for i in range(predictions.shape[0]):
            predictions[i] = model(test[i]).cpu().numpy()
            targets[i] = y_test[map_nr][i+step_size]

        print("Evaluation with MSE: ", evaluate(predictions, targets))

        return predictions, targets


#Gives the loss of a match calculated with MSE
def evaluate(predicted, target):
    loss = (np.square(target - predicted)).mean(axis=None)
    return loss

#Visualizes predictions vs targets
def visualize(predictions, targets, n_players):
    predictions = np.reshape(predictions, [len(predictions), n_players, 4])
    targets = np.reshape(targets, [len(targets), n_players, 4])
    vis.visualize_predicted_real((1, 1, 1), targets, predictions)


if __name__ == "__main__":
    n_players = 2
    step_size = 20
    hidden_size = 20

    predictions, targets = run(step_size, hidden_size, n_players)
    visualize(predictions, targets, n_players)
