import sys
import os.path
libdir = os.path.dirname(__file__)
sys.path.append(os.path.split(libdir)[0])

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import Data.DataFormatter as df
import numpy as np

import Visualize.Visualize as vis
from datetime import datetime
import Data.InterestPoint as ip


"""
Uses a grid that represents the map
Each element in the grid corresponds to a position on the map
Predicts possibiliy of there beeing a point of interest for each element in the grid
"""


class GridModel (nn.Module):

    def __init__(self, input_size, hidden_size, device):
        super(GridModel, self).__init__()
        self.hidden_size = hidden_size
        self.input_size = input_size
        self.device = device

        self.lstm = nn.LSTM(input_size, hidden_size)
        self.hidden = self.init_hidden()
        self.leakyRelu = torch.nn.LeakyReLU()
        self.linear = torch.nn.Linear(hidden_size, input_size)

    def init_hidden(self):
        return torch.randn(1, 1, self.hidden_size, device=self.device), torch.randn(1, 1, self.hidden_size, device=self.device)


    def forward(self, map):
        lstm_out, self.hidden = self.lstm(map.view(1, 1, -1), self.hidden)
        lstm_out = F.tanh(lstm_out[0][0])
        lstm_out = self.leakyRelu(lstm_out)
        lstm_out = self.linear(lstm_out)
        lstm_out = F.softmax(lstm_out)
        return lstm_out

#gets the data from dir scaled by factor and the corresponding interesting positions
#converts the numpy arrays to tensors, and reshapes each match to [timesteps, features]
def get_data(factor):
    x_train, x_test = df.get_data_from_dir("Data/Datasets/set1", 2, factor=factor)
    int_point = ip.InterestPoint([7908/factor, 1890/factor, 830/factor])
    y_train = int_point.get_most_interesting_positions(x_train)
    y_test = int_point.get_most_interesting_positions(x_test)

    for i in range(len(x_train)):
        x_train[i] = np.reshape(x_train[i], [x_train[i].shape[0], -1])
        x_train[i] = torch.from_numpy(x_train[i]).float()
        y_train[i] = np.reshape(y_train[i], [y_train[i].shape[0], -1])
        y_train[i] = torch.from_numpy(y_train[i]).float()

    for i in range(len(x_test)):
        x_test[i] = np.reshape(x_test[i], [x_test[i].shape[0], -1])
        x_test[i] = torch.from_numpy(x_test[i]).float()
        y_test[i] = np.reshape(y_test[i], [y_test[i].shape[0], -1])
        y_test[i] = torch.from_numpy(y_test[i]).float()


    return x_train, y_train, x_test, y_test

#creates a grid with the size of shape, and initializes all elements to zero
#puts the x and y positions of data as 1 in the grid
def to_grid_2d(data, shape):
    grid = torch.zeros(shape)
    for i in range(int(len(data)/4)):
        grid[int(data[(i*4)])][int(data[(i*4)+1])] = 1.0

    return grid.view(-1)


#the main training loop
def run(step_size, shape, factor):
    shape = (int(shape[0] / factor)+1, int(shape[1] / factor)+1)
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    x_train, y_train, x_test, y_test = get_data(factor)
    gridshape = list(to_grid_2d(x_train[0][0], shape).shape)[0]
    print(gridshape)
    model = GridModel(gridshape, 400, device)
    del gridshape
    model.to(device)
    date = datetime.now()

    loss_func = torch.nn.MSELoss()
    optimizer = optim.Adam(model.parameters(), lr=0.0001)

    lowest_loss = 10000.0
    for epoch in range(10):
        print("epoch %i" % epoch)
        print(datetime.now().time())
        running_loss = 0.0
        for j in range(len(x_train)):
            print("match ", j, ": ", datetime.now().time())
            data_in = x_train[j]
            data_out = y_train[j]
            #print("map")

            for i in range(len(data_in)-step_size):
                #print("timestep ", i, " of ", len(data_in)-step_size)
                optimizer.zero_grad()
                model.hidden = model.init_hidden()

                map_in = to_grid_2d(data_in[i], shape).to(device)
                target = to_grid_2d(data_out[i+step_size], shape).to(device)

                output = model(map_in)
                loss = loss_func(output, target)
                running_loss += loss

                loss.backward()
                optimizer.step()

                del map_in
                del target
                del loss

        print("loss: %f" % running_loss)
        if running_loss < lowest_loss:
            lowest_loss = running_loss
            torch.save(model.state_dict(), "Data/Models/grid_tanh_leaky_softmax_ss1_" + str(date) + ".pt")
            print("saved")


if __name__ == "__main__":
    factor = 100
    shape = (8070, 3780)
    n_players = 2
    step_size = 1
    #predictions, targets = run(step_size)
    run(step_size, shape, factor)