import sys
import os.path
libdir = os.path.dirname(__file__)
sys.path.append(os.path.split(libdir)[0])

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import Data.DataFormatter as df
import numpy as np
from sklearn import preprocessing

import Visualize.Visualize as vis
from datetime import datetime
import Data.InterestPoint as ip
import pandas as pd


"""
Tries to predict interest points directly with carthesian coordinates
Retains the hidden layer through the entire match
"""

class LstmModel1 (nn.Module):

    def __init__(self, input_size, hidden_size, device):
        super(LstmModel1, self).__init__()
        seed = 17
        np.random.seed(seed)
        torch.manual_seed(seed)

        self.hidden_size = hidden_size
        self.device = device

        self.lstm = nn.LSTM(input_size, hidden_size)
        self.hidden = self.init_hidden()
        self.leakyRelu = torch.nn.LeakyReLU()
        self.linear = torch.nn.Linear(hidden_size, input_size)

    def init_hidden(self):
        return torch.zeros(1, 1, self.hidden_size, device=self.device), torch.zeros(1, 1, self.hidden_size, device=self.device)


    def forward(self, map, len):
        lstm_out, self.hidden = self.lstm(map.view(len, 1, -1), self.hidden)
        lstm_out = self.linear(lstm_out)
        lstm_out = self.leakyRelu(lstm_out)
        lstm_out = lstm_out.view(lstm_out.shape[0], lstm_out.shape[2])
        return lstm_out

#Gets the data from directory scaled to [0, 1], and gets the corresponding most interesting positions
#reshapes the numpy arrays for each match to [timesteps, features] and converts to tensors
def get_data():
    x_train, x_test = df.get_data_from_dir("Data/Datasets/larger_set", 4, scale=True)
    int_point = ip.InterestPoint([0.98, 0.50, 0.82])
    y_train = int_point.get_most_interesting_positions_old(x_train)
    y_test = int_point.get_most_interesting_positions_old(x_test)

    for i in range(len(x_train)):
        x_train[i] = np.reshape(x_train[i], [x_train[i].shape[0], -1])
        x_train[i] = torch.from_numpy(x_train[i]).float()[:-1]
        y_train[i] = np.reshape(y_train[i], [y_train[i].shape[0], -1])
        y_train[i] = torch.from_numpy(y_train[i]).float()[1:]

    for i in range(len(x_test)):
        x_test[i] = np.reshape(x_test[i], [x_test[i].shape[0], -1])
        x_test[i] = torch.from_numpy(x_test[i]).float()
        y_test[i] = np.reshape(y_test[i], [y_test[i].shape[0], -1])
        y_test[i] = torch.from_numpy(y_test[i]).float()

    return x_train, y_train, x_test, y_test


#The main training loop
def run(step_size, hidden_size):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    print(device)
    x_train, y_train, x_test, y_test = get_data()
    model = LstmModel1(x_train[0].shape[1], hidden_size, device)
    #model.load_state_dict(torch.load("/home/petter/Bachelor/jupspec/DeepSpec/Data/Models/OldModels/transfer_test_base_2019-04-24 12:11:21.427295.pt", map_location="cpu"))
    model.to(device)
    date = datetime.now()
    save_string = "Data/Models/Misc/lstm_large_50k" + str(date)

    loss_func = torch.nn.MSELoss()
    optimizer = optim.Adam(model.parameters(), lr=0.0001)

    lowest_loss = 10000.0

    loss_history = []
    for epoch in range(50000):
        print("epoch %i" % epoch)
        print(datetime.now().time())
        running_loss = 0.0

        for j in range(len(x_train)):
            #print("match")
            data_in = x_train[j].to(device)
            data_out = y_train[j].to(device)
            #print("map")
            model.hidden = model.init_hidden()
            optimizer.zero_grad()

            output = model(data_in, data_in.shape[0])

            loss = loss_func(output, data_out)
            running_loss += loss.item()

            loss.backward()
            optimizer.step()

        running_loss = (running_loss/len(x_train))*1000
        loss_history.append(running_loss)
        print("loss: %f" % running_loss)
        if running_loss < lowest_loss:
            lowest_loss = running_loss
            torch.save(model.state_dict(), save_string + ".pt")
            print("saved")

    return loss_history, save_string


if __name__ == "__main__":
    n_players = 4
    step_size = 1
    hidden= 64
    hist, string = run(step_size, hidden)

    df = pd.DataFrame(hist)
    df.to_csv(string + "_history.txt", header=None)

    print("loss:")
    print(hist)
    vis.visualize_history(hist)

