import sys
import os.path
libdir = os.path.dirname(__file__)
sys.path.append(os.path.split(libdir)[0])

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import Data.DataFormatter as df
import numpy as np
from sklearn import preprocessing

import Visualize.Visualize as vis
from datetime import datetime
import Data.InterestPoint as ip
import pandas as pd


"""
Tries to predict interest points directly with carthesian coordinates
"""

class LstmModel2 (nn.Module):

    def __init__(self, input_size, hidden_size, device):
        super(LstmModel2, self).__init__()
        seed = 17
        np.random.seed(seed)
        torch.manual_seed(seed)

        self.hidden_size = hidden_size
        self.device = device

        self.lstm = nn.LSTM(input_size, hidden_size, num_layers=1)
        self.leakyRelu = torch.nn.LeakyReLU()
        self.linear = torch.nn.Linear(hidden_size, 3)


    def forward(self, map):
        lstm_out, hidden = self.lstm(map.view(1, 1, -1))
        #lstm_out = F.tanh(lstm_out[0][0])
        lstm_out = self.leakyRelu(lstm_out)
        lstm_out = self.linear(lstm_out)
        lstm_out = self.leakyRelu(lstm_out)
        return lstm_out


#Gets the data from directory scaled to [0, 1], and gets the corresponding most interesting positions
#reshapes the numpy arrays for each match to [timesteps, features] and converts to tensors
def get_data():
    x_train, x_test = df.get_data_from_dir("Data/Datasets/larger_set", 4, scale=True)
    int_point = ip.InterestPoint([0.98, 0.50, 0.82])
    y_train = int_point.get_most_interesting_positions(x_train)
    y_test = int_point.get_most_interesting_positions(x_test)

    for i in range(len(x_train)):
        x_train[i] = np.reshape(x_train[i], [x_train[i].shape[0], -1])
        x_train[i] = torch.from_numpy(x_train[i]).float()
        y_train[i] = np.reshape(y_train[i], [y_train[i].shape[0], -1])
        y_train[i] = torch.from_numpy(y_train[i]).float()

    for i in range(len(x_test)):
        x_test[i] = np.reshape(x_test[i], [x_test[i].shape[0], -1])
        x_test[i] = torch.from_numpy(x_test[i]).float()
        y_test[i] = np.reshape(y_test[i], [y_test[i].shape[0], -1])
        y_test[i] = torch.from_numpy(y_test[i]).float()

    return x_train, y_train, x_test, y_test


#The main training loop
def run(step_size):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    x_train, y_train, x_test, y_test = get_data()
    model = LstmModel2(x_train[0].shape[1], 64, device)
    model.to(device)
    date = datetime.now()
    save_string = "Data/Models/Misc/rnn_large_ss40_2k" + str(date)

    loss_func = torch.nn.MSELoss()
    optimizer = optim.Adam(model.parameters(), lr=0.00001, amsgrad=True)

    lowest_loss = 1000000.0
    loss_hist = []
    for epoch in range(2000):
        print("epoch %i" % epoch)
        print(datetime.now().time())
        running_loss = 0.0
        for j in range(len(x_train)):
            data_in = x_train[j].to(device)
            data_out = y_train[j].to(device)

            for i in range(len(data_in)-step_size):
                optimizer.zero_grad()

                map_in = data_in[i]
                target = data_out[i+step_size]

                output = model(map_in)
                loss = loss_func(output, target)
                running_loss += loss.item()

                loss.backward()
                optimizer.step()
            running_loss = running_loss / len(x_train[j])
        running_loss = (running_loss / len(x_train)) * 1000

        print("loss: %f" % running_loss)
        loss_hist.append(running_loss)
        if running_loss < lowest_loss:
            lowest_loss = running_loss
            torch.save(model.state_dict(), save_string + ".pt")
            print("saved")


    return loss_hist, save_string


if __name__ == "__main__":
    n_players = 4
    step_size = 40
    hist, string = run(step_size)

    df = pd.DataFrame(hist)
    df.to_csv(string + "_history.txt", header=None)

    print("loss:")
    print(hist)
    vis.visualize_history(hist)


