import sys
import os.path
libdir = os.path.dirname(__file__)
sys.path.append(os.path.split(libdir)[0])

import torch
import torch.nn as nn
import torch.optim as optim
import Data.DataFormatter as df
import numpy as np
import pandas as pd

import Visualize.Visualize as vis
from datetime import datetime
import Data.InterestPoint as ip


"""
#Uses LSTM with the last 10 timesteps to predict future interesting positions
"""

class LstmModel3 (nn.Module):

    def __init__(self, input_size, hidden_size, num_layers, step_size, device):
        super(LstmModel3, self).__init__()
        seed = 17
        np.random.seed(seed)
        torch.manual_seed(seed)

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.step_size = step_size
        self.device = device

        self.lstm = nn.LSTM(self.input_size, self.hidden_size, self.num_layers)
        self.dropout = nn.Dropout()
        self.hidden = self.init_hidden()
        self.leakyRelu = torch.nn.LeakyReLU()
        self.linear = torch.nn.Linear(hidden_size, 3)

    def init_hidden(self):
        return torch.zeros(self.num_layers, 10, self.hidden_size, device=self.device), torch.zeros(self.num_layers, 10, self.hidden_size, device=self.device)


    def forward(self, map):
        lstm_out, self.hidden = self.lstm(map.view(1, 10, -1), self.hidden)
        #self.dropout(lstm_out)
        lstm_out = self.leakyRelu(lstm_out[0][-1])
        lstm_out = self.linear(lstm_out)
        lstm_out = self.leakyRelu(lstm_out)
        return lstm_out


#Gets the data from directory scaled to [0, 1], and gets the corresponding most interesting positions
#reshapes the numpy arrays for each match to [timesteps, features] and converts to tensors
def get_data(n_players):
    x_train, x_test = df.get_data_from_dir("Data/Datasets/larger_set", n_players, scale=True)
    int_point = ip.InterestPoint([0.98, 0.50, 0.82])
    y_train = int_point.get_most_interesting_positions(x_train)
    y_test = int_point.get_most_interesting_positions(x_test)

    for i in range(len(x_train)):
        x_train[i] = np.reshape(x_train[i], [x_train[i].shape[0], -1])
        x_train[i] = torch.from_numpy(x_train[i]).float()
        y_train[i] = np.reshape(y_train[i], [y_train[i].shape[0], -1])
        y_train[i] = torch.from_numpy(y_train[i]).float()


    for i in range(len(x_test)):
        x_test[i] = np.reshape(x_test[i], [x_test[i].shape[0], -1])
        x_test[i] = torch.from_numpy(x_test[i]).float()
        y_test[i] = np.reshape(y_test[i], [y_test[i].shape[0], -1])
        y_test[i] = torch.from_numpy(y_test[i]).float()

    return x_train, y_train, x_test, y_test


#The main training loop
def run(step_size, hidden_size, num_layers, n_players):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    print(device)
    x_train, y_train, x_test, y_test = get_data(n_players)
    model = LstmModel3(x_train[0].shape[1], hidden_size, num_layers, step_size, device)
    model.to(device)
    date = datetime.now()
    save_string = "Data/Models/Misc/large_ss40_2k" + str(date)

    loss_func = torch.nn.MSELoss()
    optimizer = optim.Adam(model.parameters(), lr=0.00001, amsgrad=True)

    lowest_loss = 10000000.0
    loss_history = []

    for epoch in range(2000):
        print("epoch %i" % epoch)
        print(datetime.now().time())
        running_loss = 0.0

        for j in range(len(x_train)):
            data_in = x_train[j].to(device)
            data_out = y_train[j].to(device)

            for i in range(len(x_train[j])-(step_size*2)):
                optimizer.zero_grad()
                model.hidden = model.init_hidden()

                map = data_in[i:(i+10)]
                target = data_out[i+10+step_size]

                output = model(map)
                loss = loss_func(output, target)

                loss.backward()
                optimizer.step()

                running_loss += loss.item()

            running_loss = running_loss/len(x_train[j])
        running_loss = (running_loss/len(x_train))*1000 #*1000 to get more readable numbers (not 0.003 etc.)
        loss_history.append(running_loss)

        print("loss: %f" % running_loss)
        if running_loss < lowest_loss:
            lowest_loss = running_loss
            torch.save(model.state_dict(), save_string + ".pt")
            print("saved")


    end = datetime.now() - date
    print("Running time: ", end)
    return loss_history, save_string


if __name__ == "__main__":
    n_players = 4
    step = 40
    hidden= 64
    layers = 2
    hist, string = run(step, hidden, layers, n_players)

    df = pd.DataFrame(hist)
    df.to_csv(string + "_history.txt", header=None)

    print("loss:")
    print(hist)
    vis.visualize_history(hist)

