import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.widgets import Slider, Button, RadioButtons
import matplotlib.cm as cm
import numpy as np
from itertools import cycle

def visualizePathWithPrediction(map, real_data, pred_data):
    colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w', 'pink', 'gray']

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    def reshape_data(data):
        """ Reshapes input data to a form which matplotlib accepts
        :param data: the input data in form [[p1_x, p1_y, p1_z], [p2_x, p2_y, p2_z], ...]
        :return: reshaped data as [[[p1_x1, p1_x2, ...], [p1_y1, p1_y2, ...], [p1_z1, p1_z2, ...]], [[p2_x, ...][p2_y, ...][p2_z, ...]], ...]
        """
        num_players = len(data[0])
        reshaped_data = []

        for x in range(0, num_players):
            reshaped_data.append([[],[],[],[],[],[],[]])

        for state in data:
            for x in range(0, num_players):
                reshaped_data[x][0].append(state[x][0])
                reshaped_data[x][1].append(state[x][1])
                reshaped_data[x][2].append(state[x][2])
                #reshaped_data[x][3].append(state[x][3])
        return reshaped_data


    def set_axes():
        ax.clear()
        ax.set_xlim3d(0, map[0])
        ax.set_ylim3d(0, map[1])
        ax.set_zlim3d(0, map[2])
        ax.set_xlabel("X-axis")
        ax.set_ylabel("Y-axis")
        ax.set_zlabel("Z-axis")

    
    def show_all(reshaped_real, reshaped_pred):
        """ Visualized all the input data

        :param reshaped_real: the reshaped data for real positions
        :param reshaped_pred: the reshaped data for predicted positions

        """
        set_axes()
        ax.set_title("Predicted vs real positions")

        #rainbow = cm.rainbow(np.linspace(0, 1, len(reshaped_real[0][0])))
        #rainbow1 = cm.rainbow(np.linspace(0, 1, len(reshaped_pred[0][0])))


        counter = 0
        for player in reshaped_real:
            ax.plot(player[0], player[1], player[2], '-', c=colors[counter], label="real P%i" % (counter+1))
            #ax.scatter(player[0], player[1], player[2], c=rainbow, marker='o')
            counter +=1

        counter = 0
        for player in reshaped_pred:
            ax.plot(player[0], player[1], player[2], ':', c=colors[counter], label="pred P%i" % (counter+1))
            #ax.scatter(player[0], player[1], player[2], c=rainbow1, marker='^')
            counter +=1

        ax.legend(loc='upper left')


    def reset(event):
        """ Resets the graph to show all the data.
        :param event: required default parameter.
        """
        slider_timestep.reset()
        show_all(reshaped_real, reshaped_pred)


    def update(val):
        """ Updates the graph and show only one timestep according to slider value.
        :param val: the current slider value
        """
        timestep = slider_timestep.val
        timestep = int(timestep)
        set_axes()
        ax.set_title("Timestep %i" % timestep)

        counter = 0
        try:
            for player in reshaped_real:
                x, y, z = player[0][timestep], player[1][timestep], player[2][timestep]
                ax.scatter(x, y, z, c=colors[counter], marker='o', label="real P%i" % (counter+1))
                ax.text(x, y+0.6, z, '(%i, %i, %i)' % (x, y, z), fontsize=9)
                counter += 1
        except:
            print("An error occured, real data set is smaller than predicted. Timestep %i" % timestep)

        counter = 0
        try:
            for player in reshaped_pred:
                x, y, z = player[0][timestep], player[1][timestep], player[2][timestep]
                ax.scatter(x, y, z, c=colors[counter], marker='^', label="real P%i" % (counter+1))
                ax.text(x, y+0., z, '(%i, %i, %i)' % (x, y, z), fontsize=9)
                counter += 1
        except:
            print("An error occured, predicted data set is smaller than real. Timestep %i" % timestep)

        ax.legend(loc='upper left')


    # Reshape the input data
    reshaped_real = reshape_data(real_data)
    reshaped_pred = reshape_data(pred_data)
    
    # Calculate number of timesteps required
    num_timesteps_real = len(reshaped_real[0][0])
    num_timesteps_predicted = len(reshaped_pred[0][0])

    if num_timesteps_real >= num_timesteps_predicted:
        num_timesteps = num_timesteps_real
    else:
        num_timesteps = num_timesteps_predicted

    print("Number of timesteps: %i" % num_timesteps)
    print("Number of timesteps real: %i" % num_timesteps_real)
    print("Number of timesteps pred: %i" % num_timesteps_predicted)

    
    axcolor = 'lightgoldenrodyellow'

    # Making of the slider 
    axtimestep = plt.axes([0.25, 0.025, 0.65, 0.03], facecolor=axcolor)
    slider_timestep = Slider(axtimestep, 'Timestep', 0, num_timesteps-1, valinit=0, valstep=1)

    # Making Radio buttons
    axradio = plt.axes([0.8, 0.7, 0.1, 0.25], facecolor=axcolor)

    player_list = ['None']
    num_players2 = len(real_data[0])
    for x in range(1, num_players2+1):
        player_list.append('Player %i' % x)
    print(player_list)

    radio = RadioButtons(axradio, (player_list))

    def playerfunc(label):
        player = player_list.index('%s' % label)-1

        print(player)
        if not player == -1:
            ax.clear()
            set_axes()
            player_string = 'Player %i' % (player+1)
            ax.set_title(player_string)

            ax.plot(reshaped_real[player][0], reshaped_real[player][1], reshaped_real[player][2], '-', c=colors[player], label="real P%d" % (player+1))
            ax.plot(reshaped_pred[player][0], reshaped_pred[player][1], reshaped_pred[player][2], ':', c=colors[player], label="pred P%d" % (player+1))

            ax.legend(loc='upper left')
        else:
            show_all(reshaped_real, reshaped_pred)

    radio.on_clicked(playerfunc)


    # Making of reset button
    resetax = plt.axes([0.025, 0.025, 0.1, 0.03])
    button = Button(resetax, 'Reset', color=axcolor, hovercolor='0.975')

    
    show_all(reshaped_real, reshaped_pred)

    button.on_clicked(reset)
    slider_timestep.on_changed(update)

    plt.show()